// frustum.cpp

#include <cfloat>
#include "frustum.h"

// Plane class implementation
// --------------------------
Plane::Plane()
{
	pos = Vector3(0, 0, 0);
	normal = Vector3(0, 1, 0);
}
	
Plane::Plane(float a, float b, float c, float d)
{
	Vector3 normal = Vector3(a, b, c);
    float len = normal.Length();
    normal /= len;

	Vector3 pos = -d * len * normal;
	this->pos = pos;
	this->normal = normal;
}

Plane::Plane(const Vector3& normal, const Vector3& pos)
{
	this->normal = normal;
	this->pos = pos;
}

Vector3 Plane::GetPos() const
{
	return pos;
}
	
void Plane::SetPos(const Vector3& pos)
{
	this->pos = pos;
}

Vector3 Plane::GetNormal() const
{
	return normal;
}

void Plane::SetNormal(const Vector3& normal)
{
	this->normal = normal;
}

unsigned int Plane::ClassifyPoint (const Vector3 &point) const
{
	float result;
    Vector3 vec1 = pos;
    Vector3 direction = vec1-point;

    result = DotProduct (normal, direction);

    if (result< -0.01)
    {
        return CP_FRONT;
    }
    if (result> 0.01)
    {
        return CP_BACK;
    }
    return CP_ONPLANE;

}


// Line class implementation
// -------------------------
Line::Line()
{
	verts[0] = verts[1] = Vector3(0, 0, 0);
}

Line::Line(const Vector3& a, const Vector3& b)
{
	verts[0] = a;
	verts[1] = b;
}

Vector3 Line::GetVertex(unsigned int vert) const
{
	if (vert < 2)
	{
		return verts[vert];
	}
	return Vector3();
}

void Line::SetVertex(unsigned int vert, const Vector3& vertex)
{
	if (vert < 2)
	{
		verts[vert] = vertex;
	}
}

Ray Line::GetRay() const
{
	return Ray(verts[0], (verts[1] - verts[0]).Normalized());
}

float Line::GetLength() const
{
	return (verts[1] - verts[0]).Length();		
}

// Ray class implementation
// ------------------------
Ray::Ray()
{
	origin = Vector3(0, 0, 0);
	direction = Vector3(0, 0, 1);
}

Ray::Ray(const Vector3& origin, const Vector3& direction)
{
	this->origin = origin;
	this->direction = direction.Normalized();
}

Vector3 Ray::GetOrigin() const
{
	return origin;
}

Vector3 Ray::GetDirection() const
{
	return direction;
}

void Ray::SetOrigin(const Vector3& origin)
{
	this->origin = origin;
}

void Ray::SetDirection(const Vector3& direction)
{
	this->direction = direction.Normalized();
}

Vector3 Ray::GetPos(float dist) const
{
	return origin + (dist * direction);
}

// Triangle class implementation
// -----------------------------
Triangle::Triangle()
{
	verts[0] = verts[1] = verts[2] = Vector3(0, 0, 0);
}

Triangle::Triangle(const Vector3& a, const Vector3& b, const Vector3& c)
{
	verts[0] = a;
	verts[1] = b;
	verts[2] = c;
}

Vector3 Triangle::GetNormal() const
{
	return (CrossProduct((verts[1] - verts[0]), (verts[2] - verts[0]))).Normalized();
}

Vector3 Triangle::GetVertex(unsigned int vert) const
{
	if (vert < 3)
	{
		return verts[vert];
	}
	return Vector3();
}

void Triangle::SetVertex(unsigned int vert, const Vector3& vertex)
{
	if (vert < 3)
	{
		verts[vert] = vertex;
	}
}


Plane Triangle::GetPlane() const
{
	return Plane(GetNormal(), verts[0]);
}

Line Triangle::GetLine(unsigned int lin) const
{
	if (lin < 3)
	{
		return Line(verts[lin], verts[(lin + 1) % 3]);
	}
	return Line();
}

Plane Triangle::GetEdgePlane(unsigned int i) const
{
	if (i < 3)
	{
		Line lin = GetLine(i);
		Vector3 c = lin.GetVertex(0) + GetNormal();
		Triangle tri = Triangle(lin.GetVertex(0), c, lin.GetVertex(1));
		return tri.GetPlane();
	}
	return Plane();
}

// Quad class implementation
// -------------------------
Quad::Quad()
{
	for (unsigned int i=0; i<4; i++)
	{
		verts[i] = Vector3(0, 0, 0);
	}
}

Quad::Quad(const Vector3& a, const Vector3& b, 
		   const Vector3& c, const Vector3& d)
{
	verts[0] = a;
	verts[1] = b;
	verts[2] = c;
	verts[3] = d;
}

Vector3 Quad::GetVertex(unsigned int vert) const
{
	if (vert < 4)
	{
		return verts[vert];
	}
	return Vector3();
}

void Quad::SetVertex(unsigned int vert, const Vector3& vertex)
{
	if (vert < 4)
	{
		verts[vert] = vertex;
	}
}

Line Quad::GetLine(unsigned int lin) const
{
	if (lin < 4)
	{
		return Line(verts[lin], verts[(lin + 1) % 4]);
	}
	return Line();
}

Triangle Quad::GetTriangle(unsigned int tri) const
{
	switch (tri)
	{
		case 0:
			return Triangle(verts[0], verts[1], verts[2]);
			break;
		case 1:
			return Triangle(verts[0], verts[2], verts[3]);
			break;
		default:
			return Triangle();
			break;
	}
}

Vector3 Quad::GetNormal() const
{
	return GetTriangle(0).GetNormal();
}

Plane Quad::GetPlane() const
{
	return GetTriangle(0).GetPlane();	
}

Plane Quad::GetEdgePlane(unsigned int i) const
{
	if (i < 4)
	{
		Line lin = GetLine(i);
		Vector3 c = lin.GetVertex(0) + GetNormal();
		Triangle tri = Triangle(lin.GetVertex(0), c, lin.GetVertex(1));
		return tri.GetPlane();
	}
	return Plane();
}

bool Quad::FrustumCheck(const Box &frustum) const
{
	Plane box_planes[6];
	for(int i=0; i<6; i++) {
		box_planes[i] = frustum.GetPlane(i);
	}

	// first check the quad vertices...
	Vector3 qverts[4];
	for(int i=0; i<4; i++) {
		qverts[i] = GetVertex(i);

		bool vertex_inside = true;
		for(int j=2; j<6; j++) {
			unsigned int cp = box_planes[j].ClassifyPoint(qverts[i]);
			if(cp != CP_FRONT && cp != CP_ONPLANE) {
				vertex_inside = false;
				break;
			}
		}

		// if there is at least one vertex inside, return true
		if(vertex_inside) {
			return true;
		}
	}

	// otherwise check the intersection of each frustum quad with the quad
	for(int i=0; i<6; i++) {
		if(Intersect(frustum.GetQuad(i), *this)) {
			return true;
		}
	}

	return false;
}

// BoundingVolume class implementation
// -----------------------------------
BoundingVolume::BoundingVolume()
{
	transform = Matrix4x4::identity;
}

BoundingVolume::~BoundingVolume()
{
	for (size_t i=0; i<children.size(); i++)
	{
		delete children[i];
	}
}

// this will be overriden
bool BoundingVolume::FrustumCheck(const Box &frustum, const Matrix4x4 &transform) const {
	return true;
}

bool BoundingVolume::FrustumCheck(const Box& frustum) const
{
	if (!FrustumCheck(frustum, transform))
	{
		return false;
	}

	if (children.size() == 0)
	{
		return true;
	}

	for (unsigned int i=0; i<children.size(); i++)
	{
		if (children[i]->FrustumCheck(frustum))
		{
			return true;
		}
	}

	return false;
}

void BoundingVolume::SetTransform(const Matrix4x4& mat)
{
	this->transform = mat;
	for (unsigned int i=0; i<children.size(); i++)
	{
		children[i]->SetTransform(mat);
	}
}

void BoundingVolume::AddVolume(BoundingVolume* vol)
{
	children.push_back(vol);
}

// Tetrahedron class implementation
// --------------------------------
Tetrahedron::Tetrahedron()
{
	for (unsigned int i=0; i<4; i++)
	{
		verts[i] = Vector3(0, 0, 0);
	}
}

Tetrahedron::Tetrahedron(const Vector3& a, const Vector3& b, 
						 const Vector3& c, const Vector3& d)
{
	verts[0] = a;
	verts[1] = b;
	verts[2] = c;
	verts[3] = d;
}

Vector3 Tetrahedron::GetVertex(unsigned int vert) const
{
	if (vert < 4)
	{
		return verts[vert];
	}
	return Vector3();
}

void Tetrahedron::SetVertex(unsigned int vert, const Vector3& vertex)
{
	if (vert < 4)
	{
		verts[vert] = vertex;
	}
}

Triangle Tetrahedron::GetTriangle(unsigned int tri) const
{
	switch (tri)
	{
		case 0:
			return Triangle(verts[0], verts[1], verts[2]);
			break;
		case 1:
			return Triangle(verts[0], verts[4], verts[1]);
			break;
		case 2:
			return Triangle(verts[1], verts[4], verts[2]);
			break;
		case 3:
			return Triangle(verts[2], verts[4], verts[0]);
			break;
		default:
			return Triangle();
			break;
	}
}

bool Tetrahedron::FrustumCheck(const Box& frustum, const Matrix4x4& transform) const
{
	Tetrahedron transformed = *this;
	transformed.Transform(transform);
	
	unsigned char outside[4];
	memset(outside, 0, 4 * sizeof(unsigned char));
	
	// STEP 1: collect info about which vertex is in/outside which frustum plane
	// if any vertex is inside all frustum planes, then we return success
	for (unsigned int t_vert=0; t_vert<4; t_vert++)
	{
		for (unsigned int f_plane=0; f_plane<6; f_plane++)
		{
			if (UnderPlane(transformed.GetVertex(t_vert), frustum.GetPlane(f_plane)))
			{
				outside[t_vert] |= (1 << f_plane);
			}
		}
		if (!outside[t_vert])
		{
			// this vertex is inside the frustum
			return true;
		}
	}

	// STEP 2: if all vertices are outside *any* plane then return failure
	unsigned char all_verts = outside[0];

	for (unsigned int i=1; i<4; i++)
	{
		all_verts &= outside[i];
	}

	if (all_verts)
	{
		// all verts are outside of at least one plane
		return false;
	}

	// STEP 3: if any quad of the frustum intersects any 
	// triangle of the tetrahedron, return success
	for (unsigned int f_quad=0; f_quad<6; f_quad++)
	{
		for (unsigned int t_tri=0; t_tri<4; t_tri++)
		{
			if (Intersect(transformed.GetTriangle(t_tri), frustum.GetQuad(f_quad)))
			{
				return true;
			}
		}
	}

	// STEP 4: there are 2 cases left: 
	// 1 - the tetrahedron is completely outside the frustum
	// 2 - the tetrahedron contains the frustum
	// if any point of the frustum is outside any plane 
	// of the Tetrahedron then return failure
	for (unsigned int f_vert=0; f_vert<8; f_vert++)
	{
		for (unsigned int t_plane=0; t_plane<4; t_plane++)
		{
			if (UnderPlane(frustum.GetVertex(f_vert), transformed.GetTriangle(t_plane).GetPlane()))
			{
				return false;
			}
		}
	}

	// yup! the bounding tetrahedron *contains* the frustum :)
	return true;
}

void Tetrahedron::Transform(const Matrix4x4& mat)
{
	for (unsigned int i=0; i<4; i++)
	{
		verts[i].Transform(mat);
	}
}

// Box class implementation
// ------------------------
void Box::CalculatePlanes()
{
	for (unsigned int i=0; i<6; i++)
	{
		Quad q = GetQuad(i);
		planes[i] = Plane(q.GetNormal(), q.GetVertex(0));
	}
}

void Box::CalculateVertices()
{
	verts[0] = PointOfIntersection(planes[FP_FRONT], planes[FP_TOP], planes[FP_LEFT]);
	verts[1] = PointOfIntersection(planes[FP_FRONT], planes[FP_BOTTOM], planes[FP_LEFT]);
	verts[2] = PointOfIntersection(planes[FP_FRONT], planes[FP_BOTTOM], planes[FP_RIGHT]);
	verts[3] = PointOfIntersection(planes[FP_FRONT], planes[FP_TOP], planes[FP_RIGHT]);

	verts[4] = PointOfIntersection(planes[FP_BACK], planes[FP_TOP], planes[FP_LEFT]);
	verts[5] = PointOfIntersection(planes[FP_BACK], planes[FP_BOTTOM], planes[FP_LEFT]);
	verts[6] = PointOfIntersection(planes[FP_BACK], planes[FP_BOTTOM], planes[FP_RIGHT]);
	verts[7] = PointOfIntersection(planes[FP_BACK], planes[FP_TOP], planes[FP_RIGHT]);
}

Box::Box()
{
	for (unsigned int i=0; i<6; i++)
	{
		planes[i] = Plane();
	}

	for (unsigned int i=0; i<8; i++)
	{
		verts[i] = Vector3(0, 0, 0);
	}
}

Box::Box(const Plane& front, const Plane& back,
	const Plane& left, const Plane& right,
	const Plane& top, const Plane& bottom)
{
	planes[0] = front;
	planes[1] = back;
	planes[2] = left;
	planes[3] = right;
	planes[4] = top;
	planes[5] = bottom;

	CalculateVertices();
}

Box::Box(const Quad& front, const Quad& back)
{
	verts[0] = front.GetVertex(0);
	verts[1] = front.GetVertex(1);
	verts[2] = front.GetVertex(2);
	verts[3] = front.GetVertex(3);

	verts[4] = back.GetVertex(0);
	verts[5] = back.GetVertex(1);
	verts[6] = back.GetVertex(2);
	verts[7] = back.GetVertex(3);

	CalculatePlanes();
}

Quad Box::GetQuad(unsigned int i) const
{
	switch (i)
	{
		case FP_FRONT:
			return Quad(verts[3], verts[0], verts[1], verts[2]);
			break;
		case FP_BACK:
			return Quad(verts[4], verts[7], verts[6], verts[5]);
			break;
		case FP_LEFT:
			return Quad(verts[0], verts[4], verts[5], verts[1]);
			break;
		case FP_RIGHT:
			return Quad(verts[7], verts[3], verts[2], verts[6]);
			break;
		case FP_TOP:
			return Quad(verts[0], verts[3], verts[7], verts[4]);
			break;
		case FP_BOTTOM:
			return Quad(verts[5], verts[6], verts[2], verts[1]);
			break;
		default:
			return Quad();
			break;
	}
}

Plane Box::GetPlane(unsigned int pln) const
{
	if (pln < 6)
	{
		return planes[pln];
	}
	return Plane();
}

void Box::SetPlane(unsigned int pln, const Plane& plane)
{
	if (pln < 6)
	{
		planes[pln] = plane;
		CalculateVertices();
	}
}

Vector3 Box::GetVertex(unsigned int vert) const
{
	if (vert < 8)
	{
		return verts[vert];
	}
	return Vector3();
}

void Box::SetVertex(unsigned int vert, const Vector3& vertex)
{
	if (vert < 8)
	{
		verts[vert] = vertex;
		CalculatePlanes();
	}
}

bool Box::FrustumCheck(const Box& frustum, const Matrix4x4& transform) const
{
	Box transformed = *this;
	transformed.Transform(transform);
	
	unsigned char outside[8];
	memset(outside, 0, 8 * sizeof(unsigned char));
	
	// STEP 1: collect info about which vertex is in/outside which frustum plane
	// if any vertex is inside all frustum planes, then we return success
	for (unsigned int b_vert=0; b_vert<8; b_vert++)
	{
		for (unsigned int f_plane=0; f_plane<6; f_plane++)
		{
			if (UnderPlane(transformed.GetVertex(b_vert), frustum.GetPlane(f_plane)))
			{
				outside[b_vert] |= (1 << f_plane);
			}
		}
		if (!outside[b_vert])
		{
			// this vertex is inside the frustum
			return true;
		}
	}

	// STEP 2: if all vertices are outside *any* plane then return failure
	unsigned char all_verts = outside[0];

	for (unsigned int i=1; i<8; i++)
	{
		all_verts &= outside[i];
	}

	if (all_verts)
	{
		// all verts are outside of at least one plane
		return false;
	}

	// STEP 3: if any quad of the frustum intersects any quad of the box, return success
	for (unsigned int f_quad=0; f_quad<6; f_quad++)
	{
		for (unsigned int b_quad=0; b_quad<6; b_quad++)
		{
			if (Intersect(frustum.GetQuad(f_quad), transformed.GetQuad(b_quad)))
			{
				return true;
			}
		}
	}

	// STEP 4: there are 2 cases left: 
	// 1 - the box is completely outside the frustum
	// 2 - the box contains the frustum
	// if any point of the frustum is outside any plane 
	// of the box then return failure
	for (unsigned int f_vert=0; f_vert<8; f_vert++)
	{
		for (unsigned int b_plane=0; b_plane<6; b_plane++)
		{
			if (UnderPlane(frustum.GetVertex(f_vert), transformed.GetPlane(b_plane)))
			{
				return false;
			}
		}
	}

	// yup! the bounding box *contains* the frustum :)
	return true;
}

void Box::Transform(const Matrix4x4& mat)
{
	for (unsigned int i=0; i<8; i++)
	{
		verts[i] = verts[i].Transformed(mat);
	}

	CalculatePlanes();
}

unsigned int Box::ClassifyWithPlane (const Plane &plane) const
{
	unsigned int front=0, back=0;

	for (int i = 0; i < 8; i ++)
	{
		unsigned int clid = plane.ClassifyPoint (verts[i]);
		switch (clid)
		{
		case CP_FRONT: front++; break;
		case CP_BACK: back++; break;
		case CP_ONPLANE: front++; back++; break;
		}
	}

	if ((front == 8) && (back == 8))
	{
		return CP_ONPLANE;
	}
	if (front == 8)
	{
		return CP_FRONT;
	}
	if (back == 8)
	{
		return CP_BACK;
	}
	return CP_SPLIT;
}

Vector3 Box::GetViewer() const {
	Ray r = RayOfIntersection(GetPlane(FP_TOP), GetPlane(FP_BOTTOM));
	return PointOfIntersection(r, GetPlane(FP_RIGHT));
}


// Sphere class implementation
// ---------------------------
Sphere::Sphere()
{
	radius = 0.0f;
	center = Vector3(0, 0, 0);
}

Sphere::Sphere(const Vector3 center, float radius)
{
	this->center = center;
	this->radius = radius;
}

float Sphere::GetRadius() const
{
	return radius;
}

Vector3 Sphere::GetCenter() const
{
	return center;
}

void Sphere::SetRadius(float radius)
{
	this->radius = radius;
}

void Sphere::SetCenter(const Vector3& center)
{
	this->center = center;
}

bool Sphere::FrustumCheck(const Box& frustum, const Matrix4x4& transform) const
{
	Sphere transformed = *this;
	transformed.Transform(transform);
	
	for (unsigned int i=0; i<6; i++)
	{
		if (UnderPlane(transformed, frustum.GetPlane(i)))
		{
			return false;
		}
	}
	return true;
}

void Sphere::Transform(const Matrix4x4& mat)
{
	// transform radius
	Vector3 v = center + Vector3(radius, 0, 0);
	v.Transform(mat);

	// transform center
	center.Transform(mat);

	radius = (v - center).Length();
}

// utility functions
// -----------------
bool UnderPlane(const Vector3& point, const Plane& plane)
{
	if (DotProduct(point - plane.GetPos(), plane.GetNormal()) < 0.0f)
	{
		return true;
	}
	return false;
}

bool UnderPlane(const Triangle& triangle, const Plane& plane)
{
	for (unsigned int i=0; i<3; i++)
	{
		if (!UnderPlane(triangle.GetVertex(i), plane))
		{
			return false;
		}
	}
	return true;
}

bool UnderPlane(const Quad& quad, const Plane& plane)
{
	for (unsigned int i=0; i<4; i++)
	{
		if (!UnderPlane(quad.GetVertex(i), plane))
		{
			return false;
		}
	}
	return true;
}

bool UnderPlane(const Tetrahedron& tetrahedron, const Plane& plane)
{
	for (unsigned int i=0; i<4; i++)
	{
		if (!UnderPlane(tetrahedron.GetVertex(i), plane))
		{
			return false;
		}
	}
	return true;
}

bool UnderPlane(const Box& box, const Plane& plane)
{
	for (unsigned int i=0; i<8; i++)
	{
		if (!UnderPlane(box.GetVertex(i), plane))
		{
			return false;
		}
	}
	return true;
}

bool UnderPlane(const Sphere& sphere, const Plane& plane)
{
	float dist = DistanceFromPlane(sphere.GetCenter(), plane);

	if (dist + sphere.GetRadius() < 0)
	{
		return true;
	}

	return false;
}

bool Intersect(const Line& line, const Triangle& triangle)
{
	Plane pl = triangle.GetPlane();
	if (!(UnderPlane(line.GetVertex(0), pl) ^ UnderPlane(line.GetVertex(1), pl)))
	{
		// both of the line's vertices are in the same side of the
		// plane - no intersection
		return false;
	}

	// get instersection with plane
	Vector3 isc = PointOfIntersection(line.GetRay(), pl);

	// is the intersection inside the triangle?
	for (unsigned int i=0; i<3; i++)
	{
		if (UnderPlane(isc, triangle.GetEdgePlane(i)))
		{
			return false;
		}
	}

	return true;
}

bool Intersect(const Triangle& a, const Triangle& b)
{
	for (unsigned int i=0; i<3; i++)
	{
		if (Intersect(a.GetLine(i), b))
		{
			return true;
		}
	}

	for (unsigned int i=0; i<3; i++)
	{
		if (Intersect(b.GetLine(i), a))
		{
			return true;
		}
	}

	return false;
}

bool Intersect(const Triangle& triangle, const Tetrahedron& tetrahedron)
{
	for (int i=0; i<4; i++)
	{
		Triangle tetr_tri = tetrahedron.GetTriangle(i);
		if (UnderPlane(triangle, tetr_tri.GetPlane()))
		{
			return false;
		}

		if (Intersect(triangle, tetr_tri))
		{
			return true;
		}
	}

	// triangle is contained inside the tetrahedron
	return true;
}

// TODO: implement quad functions natively, if the bottlenech is here (unlikely)
bool Intersect(const Line& line, const Quad& quad)
{
	for (int i=0; i<2; i++)
	{
		if (Intersect(line, quad.GetTriangle(i)))
		{
			return true;
		}
	}

	return false;
}

bool Intersect(const Triangle& triangle, const Quad& quad)
{
	for (int i=0; i<2; i++)
	{
		if (Intersect(triangle, quad.GetTriangle(i)))
		{
			return true;
		}
	}

	return false;
}

bool Intersect(const Quad& a, const Quad& b)
{
	for (int i=0; i<2; i++)
	{
		if (Intersect(a.GetTriangle(i), b))
		{
			return true;
		}
	}

	return false;
}

bool Intersect(const Line& l, const Plane& p)
{
	bool a = UnderPlane(l.GetVertex(0), p);
	bool b = UnderPlane(l.GetVertex(1), p);

	if (a ^ b)
	{
		return true;
	}

	return false;
}

Ray RayOfIntersection(const Plane& a, const Plane& b)
{
	Vector3 t = CrossProduct(a.GetNormal(), b.GetNormal());
	if(t.Length() < XSmallNumber) {
		return Ray();
	}

	t = CrossProduct(t, a.GetNormal());
	Ray r(a.GetPos(), t);

	Vector3 point = PointOfIntersection(r, b);
	Vector3 dir = CrossProduct(a.GetNormal(), b.GetNormal());
	return Ray(point, dir);
}

Vector3 PointOfIntersection(const Ray& ray, const Plane& plane)
{
	float dot2 = DotProduct(ray.GetDirection(), plane.GetNormal());
	if (fabsf(dot2) < 0.0000000001f)
	{
		// oops! something went wrong
		return Vector3();
	}

	Vector3 X = ray.GetOrigin() - plane.GetPos();
	float dot1 = DotProduct(X, plane.GetNormal());
	float t = -(dot1 / dot2);

	return ray.GetPos(t);
}

Vector3 PointOfIntersection(const Plane& a, const Plane& b, const Plane& c)
{
	// get the intersection of a and b as a ray
	Ray ab = RayOfIntersection(a, b);

	return PointOfIntersection(ab, c);
}

float DistanceFromPlane(const Vector3& point, const Plane& plane)
{
	return DotProduct(point - plane.GetPos(), plane.GetNormal());
}

Vector3 ProjectToPlane(const Vector3& point, const Plane& plane)
{
	return point - (DistanceFromPlane(point, plane) * plane.GetNormal());
}


Plane FrustumPlane (const Matrix4x4 &mat, int plane)
{
    int i = plane / 2;
    int neg = plane % 2;

    float a,b,c,d;

    a = mat [3][0] + (neg ? -mat[i][0] : mat [i][0]);
    b = mat [3][1] + (neg ? -mat[i][1] : mat [i][1]);
    c = mat [3][2] + (neg ? -mat[i][2] : mat [i][2]);
    d = mat [3][3] + (neg ? -mat[i][3] : mat [i][3]);

    float len = Vector3 (a, b, c).Length();
    a /= len;
    b /= len;
    c /= len;
    d /= len;

    Plane p = Plane (a,b,c,d);
    return p;
}

// Viewing frustum 2
Box CalculateViewFrustum(const Matrix4x4 &view, const Matrix4x4 &proj)
{
    Matrix4x4 comboMatrix;
    comboMatrix = proj * view;

    Plane p_planes_0 = FrustumPlane (comboMatrix, 0);
    Plane p_planes_1 = FrustumPlane (comboMatrix, 1);
    Plane p_planes_2 = FrustumPlane (comboMatrix, 2);
    Plane p_planes_3 = FrustumPlane (comboMatrix, 3);
    Plane p_planes_4 = FrustumPlane (comboMatrix, 4);
    Plane p_planes_5 = FrustumPlane (comboMatrix, 5);

    Box b = Box (p_planes_4, p_planes_5, p_planes_0, p_planes_1, p_planes_3, p_planes_2);

	return b;
}

Box ReduceFrustum(const Quad &portal, const Box &frustum)
{
	Plane new_planes[4];
	Plane far_plane = frustum.GetPlane(FP_BACK);
	Plane near_plane = frustum.GetPlane(FP_FRONT);
	Vector3 view_point = frustum.GetViewer();

	Vector3 up[] = {
		frustum.GetVertex(4) - frustum.GetVertex(5),
		frustum.GetVertex(5) - frustum.GetVertex(4),
		frustum.GetVertex(7) - frustum.GetVertex(4),
		frustum.GetVertex(4) - frustum.GetVertex(7)
	};

	// for each plane of the frustum
	for(int i=0; i<4; i++) {
		int pidx = i + 2;	// don't get front/back
		
		Plane fplane = frustum.GetPlane(pidx);

		/* XXX: The opposing plane determination is VERY dependent on the order
		 * of the FrustumPlane enumerations.
		 */
		Plane opp_plane = frustum.GetPlane(pidx % 2 == 0 ? pidx + 1 : pidx - 1);

		float max_dist = 0.0;
		Vector3 best_vert;
		
		for(int j=0; j<4; j++) {
			Vector3 v = portal.GetVertex(j);

			if(UnderPlane(v, fplane)) {
				max_dist = -1.0;	// signify that we found the plane to be the same, so fuck off
				break;
			} else {
				/* the projection to the far plane is essential, since we need essentially
				 * to use the "screen space extends" of the portal in our distance measurement.
				 *
				 * EURIKA! den theloume parallel projection sto far plane... theloume perspective
				 * projection.
				 * TODO: bres tin koryfi tis pyramidas, kai xrisimopoiise to vector (v - top)
				 * gia na kaneis project (extrude) sto bathos.
				 */
				Vector3 pv = ProjectToPlane(v, far_plane);
				float dist = DistanceFromPlane(pv, opp_plane);

				if(dist > max_dist) {
					max_dist = dist;
					best_vert = v;
				}
			}
		}

		if(max_dist < 0.0) {
			new_planes[i] = fplane;
		} else {
			// construct the plane from the closest vertex found
			Vector3 v1 = best_vert - view_point;
			Vector3 v2 = (best_vert + up[i]) - view_point;

			new_planes[i].SetPos(view_point);
			new_planes[i].SetNormal(CrossProduct(v2, v1));
		}
	}

	/* XXX: possibly similar bug... replace closest to most negative */
	float min_dist = FLT_MAX;
	Vector3 closest_vert;
	for(int i=0; i<4; i++) {
		Vector3 v = portal.GetVertex(i);
		float dist = DistanceFromPlane(v, near_plane);
		if(dist < 0.0) {
			min_dist = -1.0;
			break;
		}
		if(dist < min_dist) {
			min_dist = dist;
			closest_vert = v;
		}
	}

	return Box(min_dist < 0.0 ? near_plane : Plane(near_plane.GetNormal(), closest_vert), frustum.GetPlane(FP_BACK),
			new_planes[0], new_planes[1], new_planes[2], new_planes[3]);
}
