#include <iostream>
#include <cstdlib>
#include <cassert>
#include <cfloat>
#include <GL/glut.h>
#include <GL/glu.h>
#include "frustum.h"
#include "math3d/math3d.h"

bool init(void);
void update(void);

bool frustum_check(const Quad &q, const Box &b);
Box reduce_frustum(const Box &b, const Quad &q);

void draw_box(const Box &b, float r = 0, float g = 1, float b = 0, float sz = 1.0);
void draw_quad(const Quad &q, bool active);
void rotate_box(Box *b, const Quaternion &q);
void rotate_quad(Quad *b, const Quaternion &q);
void draw_grid(float spacing, float sz);
void clean_up(void);
void key_handler(unsigned char key, int x, int y);
void skey_handler(int key, int x, int y);
void mouse_handler(int bn, int state, int x, int y);
void motion_handler(int x, int y);

static float cam_dist_fact = 1.0;
static float cam_rot_y = 200.0;
static float cam_pitch = 30.0;

Box frustum;
Quad portal;

int main(void) {
	if(!init()) {
		return EXIT_FAILURE;
	}

	glutMainLoop();
	return 0;
}

bool init(void) {
	char *fake_argv[] = {"./foo", 0};
	int fake_argc = 1;
	
	glutInit(&fake_argc, fake_argv);
	glutInitWindowSize(800, 600);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
	glutCreateWindow("foo");

	glutKeyboardFunc(key_handler);
	glutSpecialFunc(skey_handler);
	glutMouseFunc(mouse_handler);
	glutMotionFunc(motion_handler);
	glutDisplayFunc(update);
	atexit(clean_up);

	glEnable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);

	glMatrixMode(GL_PROJECTION);
	gluPerspective(45.0, 1.3333333, 1.0, 1000.0);
	glMatrixMode(GL_MODELVIEW);

	glClearColor(0.1, 0.1, 0.1, 0);
	glClearDepth(1.0);

	frustum = Box(Quad(Vector3(-1, 1, 0), Vector3(-1, -1, 0), Vector3(1, -1, 0), Vector3(1, 1, 0)),
			Quad(Vector3(-3, 3, 4), Vector3(-3, -3, 4), Vector3(3, -3, 4), Vector3(3, 3, 4)));

	portal = Quad(Vector3(-0.5, 0.5, 2), Vector3(0.5, 0.5, 2), Vector3(0.5, -0.5, 2), Vector3(-0.5, -0.5, 2));

	return true;
}

void update(void) {
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0, 0, -10 * cam_dist_fact);
	glRotatef(cam_pitch, 1, 0, 0);
	glRotatef(cam_rot_y, 0, 1, 0);

	draw_grid(2.0, 100.0);

	draw_box(frustum);

	bool in_frust = portal.FrustumCheck(frustum);
	draw_quad(portal, in_frust);

	if(in_frust) {
		draw_box(ReduceFrustum(portal, frustum), 1, 1, 0, 2);
	}

	glutSwapBuffers();
}

void draw_box(const Box &b, float cr, float cg, float cb, float sz) {
	glLineWidth(sz);
	glDisable(GL_LIGHTING);
	glBegin(GL_LINES);
	glColor3f(cr, cg, cb);
	
	for(int i=0; i<4; i++) {
		Vector3 front = b.GetVertex(i);
		Vector3 back = b.GetVertex(i + 4);
		Vector3 front_next = b.GetVertex((i + 1) % 4);
		Vector3 back_next = b.GetVertex((i + 1) % 4 + 4);
		
		glVertex3f(front.x, front.y, front.z);
		glVertex3f(front_next.x, front_next.y, front_next.z);
		glVertex3f(back.x, back.y, back.z);
		glVertex3f(back_next.x, back_next.y, back_next.z);
		glVertex3f(front.x, front.y, front.z);
		glVertex3f(back.x, back.y, back.z);
	}

	glEnd();
	glEnable(GL_LIGHTING);
	glLineWidth(1.0);
}

void draw_quad(const Quad &q, bool active) {
	glDisable(GL_LIGHTING);

	glBegin(GL_LINES);
	if(active) {
		glColor3f(1, 0, 0);
	} else {
		glColor3f(0, 0, 1);
	}

	for(int i=0; i<4; i++) {
		Vector3 v = q.GetVertex(i);
		Vector3 n = q.GetVertex((i + 1) % 4);

		glVertex3f(v.x, v.y, v.z);
		glVertex3f(n.x, n.y, n.z);
	}
	glEnd();

	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE);
	glBegin(GL_QUADS);

	if(active) {
		glColor3f(0.4, 0, 0);
	} else {
		glColor3f(0, 0, 0.4);
	}
	for(int i=0; i<4; i++) {
		Vector3 v = q.GetVertex(i);
		glVertex3f(v.x, v.y, v.z);
	}
	glEnd();
	glDisable(GL_BLEND);

	glEnable(GL_LIGHTING);
}

void rotate_box(Box *b, const Quaternion &q) {
	for(int i=0; i<16; i++) {
		Vector3 v = b->GetVertex(i);
		v.Transform(q);
		b->SetVertex(i, v);
	}
}

void rotate_quad(Quad *b, const Quaternion &q) {
	Vector3 verts[4];
	Vector3 pivot;

	for(int i=0; i<4; i++) {
		verts[i] = b->GetVertex(i);
		pivot += verts[i];
	}
	pivot /= 4.0;

	for(int i=0; i<4; i++) {
		Matrix4x4 xform;
		xform.Translate(pivot);
		xform = xform * (Matrix4x4)q.GetRotationMatrix();
		xform.Translate(-pivot);
		verts[i].Transform(xform);
		b->SetVertex(i, verts[i]);
	}
}

void draw_grid(float spacing, float sz) {
	float hsz = sz * 0.5;
	int lines = (int)(hsz / spacing) + 1;

	glDisable(GL_LIGHTING);
	glBegin(GL_LINES);
	glColor3f(0.4, 0.4, 0.4);
	for(int i=1; i<lines; i++) {
		glVertex3f(i * spacing, 0, -hsz);
		glVertex3f(i * spacing, 0, hsz);
		glVertex3f(-i * spacing, 0, -hsz);
		glVertex3f(-i * spacing, 0, hsz);
		glVertex3f(-hsz, 0, i * spacing);
		glVertex3f(hsz, 0, i * spacing);
		glVertex3f(-hsz, 0, -i * spacing);
		glVertex3f(hsz, 0, -i * spacing);
	}
	glEnd();

	glLineWidth(2.0);
	glBegin(GL_LINES);
	glColor3f(0.5, 0.5, 0.5);
	glVertex3f(-hsz, 0, 0);
	glVertex3f(hsz, 0, 0);
	glVertex3f(0, 0, -hsz);
	glVertex3f(0, 0, hsz);
	glEnd();
	glLineWidth(1.0);
	glEnable(GL_LIGHTING);
}

void clean_up(void) {
}

void key_handler(unsigned char key, int x, int y) {
	int i;

	switch(key) {
	case 27:
		exit(0);
	
	case 'm':
		for(i=0; i<13; i++) {
			rotate_box(&frustum, Quaternion(Vector3(0, 1, 0), -DEG_TO_RAD(4)));
		}
		for(i=0; i<10; i++) {
			rotate_box(&frustum, Quaternion(Vector3(1, 0, 0), DEG_TO_RAD(4)));
		}
		break;

	case 'q':
		cam_dist_fact -= 0.1;
		break;

	case 'z':
		cam_dist_fact += 0.1;
		break;

	case 'a':
		rotate_box(&frustum, Quaternion(Vector3(0, 1, 0), DEG_TO_RAD(4)));
		break;

	case 'd':
		rotate_box(&frustum, Quaternion(Vector3(0, 1, 0), -DEG_TO_RAD(4)));
		break;

	case 'w':
		rotate_box(&frustum, Quaternion(Vector3(1, 0, 0), -DEG_TO_RAD(4)));
		break;

	case 's':
		rotate_box(&frustum, Quaternion(Vector3(1, 0, 0), DEG_TO_RAD(4)));
		break;


	case 'j':
		rotate_quad(&portal, Quaternion(Vector3(0, 1, 0), DEG_TO_RAD(4)));
		break;

	case 'l':
		rotate_quad(&portal, Quaternion(Vector3(0, 1, 0), -DEG_TO_RAD(4)));
		break;

	case 'i':
		rotate_quad(&portal, Quaternion(Vector3(1, 0, 0), -DEG_TO_RAD(4)));
		break;

	case 'k':
		rotate_quad(&portal, Quaternion(Vector3(1, 0, 0), DEG_TO_RAD(4)));
		break;

	default:
		break;
	}

	glutPostRedisplay();
}

void skey_handler(int key, int x, int y) {
	switch(key) {
	case GLUT_KEY_LEFT:
		cam_rot_y += 4;
		break;

	case GLUT_KEY_RIGHT:
		cam_rot_y -= 4;
		break;

	default:
		break;
	}

	glutPostRedisplay();
}


static int prev_x, prev_y;
static int enable_rot = 0;

void mouse_handler(int bn, int state, int x, int y)
{
	if(bn == GLUT_LEFT_BUTTON) {
		prev_x = x;
		prev_y = y;
		enable_rot = state == GLUT_DOWN ? 1 : 0;
	}
}


void motion_handler(int x, int y)
{
	int dx, dy;

	if(enable_rot) {
		dx = x - prev_x;
		dy = y - prev_y;

		cam_rot_y += (float)dx * 0.2;
		cam_pitch += (float)dy * 0.1;

		prev_x = x;
		prev_y = y;

		glutPostRedisplay();
	}
}
