#include "quaternion.h"
#include "matrix.h"

using namespace std; // for C++Builder compatibility

Quaternion::Quaternion() {
	s = 1.0;
	v.x = v.y = v.z = 0.0;
}

Quaternion::Quaternion(float s, float x, float y, float z) {
	v.x = x;
	v.y = y;
	v.z = z;
	this->s = s;
}

Quaternion::Quaternion(float s, const Vector3 &v) {
	this->s = s;
	this->v = v;
}

Quaternion::Quaternion(const Vector3 &axis, float angle) {
	SetRotation(axis, angle);
}

Quaternion Quaternion::operator +(const Quaternion &quat) const {
	return Quaternion(s + quat.s, v + quat.v);
}

Quaternion Quaternion::operator -(const Quaternion &quat) const {
	return Quaternion(s - quat.s, v - quat.v);
}

Quaternion Quaternion::operator -() const {
	return Quaternion(-s, -v);
}

/** Quaternion Multiplication:
 * Q1*Q2 = [s1*s2 - v1.v2,  s1*v2 + s2*v1 + v1(x)v2]
 */
Quaternion Quaternion::operator *(const Quaternion &quat) const {
	Quaternion newq;	
	newq.s = s * quat.s - DotProduct(v, quat.v);
	newq.v = quat.v * s + v * quat.s + CrossProduct(v, quat.v);	
	return newq;
}

void Quaternion::operator +=(const Quaternion &quat) {
	*this = Quaternion(s + quat.s, v + quat.v);
}

void Quaternion::operator -=(const Quaternion &quat) {
	*this = Quaternion(s - quat.s, v - quat.v);
}

void Quaternion::operator *=(const Quaternion &quat) {
	*this = *this * quat;
}

void Quaternion::ResetIdentity() {
	s = 1.0;
	v.x = v.y = v.z = 0.0;
}

Quaternion Quaternion::Conjugate() const {
	return Quaternion(s, -v);
}

float Quaternion::Length() const {
	return (float)sqrt(v.x*v.x + v.y*v.y + v.z*v.z + s*s);
}

/** Q * ~Q = ||Q||^2 */
float Quaternion::LengthSq() const {
	return v.x*v.x + v.y*v.y + v.z*v.z + s*s;
}

void Quaternion::Normalize() {
	float len = (float)sqrt(v.x*v.x + v.y*v.y + v.z*v.z + s*s);
	v.x /= len;
	v.y /= len;
	v.z /= len;
	s /= len;
}

Quaternion Quaternion::Normalized() const {
	Quaternion nq = *this;
	float len = (float)sqrt(v.x*v.x + v.y*v.y + v.z*v.z + s*s);
	nq.v.x /= len;
	nq.v.y /= len;
	nq.v.z /= len;
	nq.s /= len;
	return nq;
}

/** Quaternion Inversion: Q^-1 = ~Q / ||Q||^2 */
Quaternion Quaternion::Inverse() const {
	Quaternion inv = Conjugate();
	float lensq = LengthSq();
	inv.v /= lensq;
	inv.s /= lensq;

	return inv;
}


void Quaternion::SetRotation(const Vector3 &axis, float angle) {
	float half_angle = angle / 2.0f;
	s = cos(half_angle);
	v = axis * sin(half_angle);
}

void Quaternion::Rotate(const Vector3 &axis, float angle) {
	Quaternion q;
	float half_angle = angle / 2.0f;
	q.s = cos(half_angle);
	q.v = axis * sin(half_angle);

	*this *= q;
}

void Quaternion::Rotate(const Quaternion &q) {
	*this = q * *this * q.Conjugate();
}

Matrix3x3 Quaternion::GetRotationMatrix() const {
	return Matrix3x3(	1.0f - 2.0f * v.y*v.y - 2.0f * v.z*v.z,	2.0f * v.x  * v.y + 2.0f * s * v.z,		2.0f * v.z * v.x - 2.0f * s * v.y,
						2.0f * v.x  * v.y - 2.0f * s * v.z,		1.0f - 2.0f * v.x*v.x - 2.0f * v.z*v.z,	2.0f * v.y * v.z + 2.0f * s * v.x,
						2.0f * v.z  * v.x + 2.0f * s * v.y,		2.0f * v.y  * v.z - 2.0f * s * v.x,		1.0f - 2.0f * v.x*v.x - 2.0f * v.y*v.y);
}


/** Spherical linear interpolation (slerp) */
Quaternion Slerp(const Quaternion &q1, const Quaternion &q2, float t) {
	float dot = DotProduct(Vector4(q1.s, q1.v.x, q1.v.y, q1.v.z), Vector4(q2.s, q2.v.x, q2.v.y, q2.v.z));
	
	if(fabs(1.0 - dot) < XSmallNumber) return q1;	// avoids divisions by zero later on if q1 == q2

        // avoid out of range error
        if (dot < -1) dot = -0.99999f;
        if (dot > 1) dot = 0.99999f;

	if(dot >= 0.0f) {
		float angle = acos(dot);
		float coef1 = (angle * sin(1.0f - t)) / sin(angle);
		float coef2 = sin(angle * t) / sin(angle);
		return Quaternion(q1.s * coef1 + q2.s * coef2, q1.v * coef1 + q2.v * coef2).Normalized();
	} else {
		float angle = acos(-dot);
		float coef1 = (angle * sin(1.0f - t)) / sin(angle);
		float coef2 = sin(angle * t) / sin(angle);
		return Quaternion(q2.s * coef1 + q1.s * coef2, q2.v * coef1 + q1.v * coef2).Normalized();
	}
}
	


std::ostream &operator <<(std::ostream &out, const Quaternion &q) {
	out << "(" << q.s << ", " << q.v << ")";
	return out;
}
